@@ -1,8 +1,8 @@
 package HTML::Template::Compiled;
-# $Id: Compiled.pm,v 1.330 2007/11/12 19:53:16 tinita Exp $
+# $Id: Compiled.pm 1070 2008-07-26 11:05:31Z tinita $
 # doesn't work with make tardist
 #our $VERSION = ($version_pod =~ m/^\$VERSION = "(\d+(?:\.\d+)+)"/m) ? $1 : "0.01";
-our $VERSION = "0.90";
+our $VERSION = "0.92";
 use Data::Dumper;
 BEGIN {
 use constant D => $ENV{HTC_DEBUG} || 0;
@@ -43,6 +43,9 @@ use constant CHECKED  => 1;
 use constant LMTIME   => 2;
 use constant LCHECKED => 3;
 
+our $DEBUG = 0;
+our $LAST_EXCEPTION;
+
 # options / object attributes
 use constant PARAM => 0;
 
@@ -53,11 +56,12 @@ BEGIN {
           file_cache cache_dir cache search_path
           loop_context case_sensitive global_vars
           default_path
-          debug perl out_fh default_escape
+          debug debug_file objects perl out_fh default_escape
           filter formatter
-          globalstack use_query parser compiler includes
+          globalstack use_query parse_tree parser compiler includes
           plugins open_mode
         )
+          #use_expressions
     );
 
     for my $i ( 1 .. $#map ) {
@@ -184,8 +188,8 @@ sub new_file {
     $self->init_cache(\%args);
     #$self->set_cache_dir( $args{cache_dir} );
     $self->set_path( $args{path} );
-    if (my $t = $self->from_cache()) {
-        $t->init_includes;
+    if (my $t = $self->from_cache(\%args)) {
+        $t->init_includes();
         return $t;
     }
     $self->init(%args);
@@ -209,7 +213,7 @@ sub new_filehandle {
     $self->init_cache(\%args);
     #$self->set_cache_dir( $args{cache_dir} );
     $self->set_path( $args{path} );
-    if (my $t = $self->from_cache()) {
+    if (my $t = $self->from_cache(\%args)) {
         return $t;
     }
     $self->init(%args);
@@ -248,7 +252,7 @@ sub new_scalar_ref {
     $self->set_filename($md5);
     D && $self->log("md5: $md5");
     $self->set_path( $args{path} );
-    if (my $t = $self->from_cache()) {
+    if (my $t = $self->from_cache(\%args)) {
         return $t;
     }
     $self->init(%args);
@@ -270,6 +274,7 @@ sub init_includes {
             $htc = $self->new_from_object($path,$filename,$fullpath,$cache);
         }
         $includes->{$fullpath}->[2] = $htc;
+        $includes->{$fullpath}->[2]->set_plugins($self->get_plugins);
     }
 }
 
@@ -318,10 +323,12 @@ sub from_scratch {
 }
 
 sub from_cache {
-    my ($self) = @_;
+    my ($self, $args) = @_;
     my $t;
     D && $self->log( "from_cache() filename=" . $self->get_filename );
 
+    $args ||= {};
+    my $plug = $args->{plugin} || [];
     # try to get memory cache
     if ( $self->get_cache ) {
         my $dir = $self->get_cache_dir;
@@ -329,6 +336,7 @@ sub from_cache {
         my $fname  = $self->get_filename;
         $t = $self->from_mem_cache($dir,$fname);
         if ($t) {
+            $t->set_plugins($plug) if @$plug;
             return $t;
         }
     }
@@ -343,6 +351,7 @@ sub from_cache {
         my $dir     = $self->get_cache_dir;
         $t = $self->from_file_cache($dir, $file);
         if ($t) {
+            $t->set_plugins($plug) if @$plug;
             return $t;
         }
     }
@@ -395,6 +404,15 @@ sub from_cache {
         D && $self->log( "add_mem_cache $fname" );
         my $clone = $self->clone;
         $clone->clear_params();
+        my @plugs = @{ $self->get_plugins || [] };
+        for my $i (0 .. $#plugs) {
+            if (ref $plugs[$i]) {
+                if ($plugs[$i]->can('serialize')) {
+                    $plugs[$i] = $plugs[$i]->serialize();
+                }
+            }
+        }
+        $clone->set_plugins(\@plugs);
         $cache->{$dir}->{$fname} = $clone;
         $times->{$dir}->{$fname} = \%times;
     }
@@ -546,7 +564,7 @@ sub add_file_cache {
       . ( join ', ', map { $self->quote_file($_) } @$path )
       . ']';
     my $isScalar = $self->get_scalar ? 1 : 0;
-    my $query_info = $self->get_use_query;
+    my $query_info = $self->get_parse_tree;
     $query_info = Data::Dumper->Dump([$query_info], ['query_info']);
     my $parser =$self->get_parser;
     $parser = Data::Dumper->Dump([\$parser], ['parser']);
@@ -561,6 +579,10 @@ sub add_file_cache {
     #$includes_to_string =~ s/\$includes = //;
     my $search_path = $self->get_search_path || 0;
     my $gl = $self->get_global_vars;
+    my $debug_file = $self->get_debug_file;
+    $debug_file =~ tr/a-z0,//cd;
+    my $use_objects = $self->get_objects;
+    $use_objects =~ tr/a-z0//cd;
     my $plugins = $self->get_plugins;
     my $plugin_dump = Data::Dumper->Dump([\@$plugins], ['plugin_dump']);
     my $file_args = $isScalar
@@ -603,6 +625,8 @@ $file_args
     use_query => \$query_info,
     parser => \$parser,
     global_vars => $gl,
+    debug_file => '$debug_file',
+    objects => '$use_objects',
     includes => \$includes,
     search_path_on_include => $search_path,
     plugin => \$plugin_dump,
@@ -650,6 +674,7 @@ sub include_file {
         eval {
             $cache = Storable::retrieve($req);
         };
+        #warn __PACKAGE__.':'.__LINE__.": error? $@\n";
         return if $@;
         my $cached_version = $cache->{version};
         $t = $cache->{htc};
@@ -658,11 +683,11 @@ sub include_file {
             return;
         }
         my $plug = $t->get_plugins || [];
-        $t->init_plugins(@$plug);
-		$t->get_cache and $t->add_mem_cache(
-			checked => $cache->{times}->{checked},
-			mtime   => $cache->{times}->{mtime},
-		);
+        $t->get_cache and $t->add_mem_cache(
+            checked => $cache->{times}->{checked},
+            mtime   => $cache->{times}->{mtime},
+        );
+        $t->init_plugins($plug);
     }
     else {
     if ($UNTAINT) {
@@ -815,13 +840,17 @@ sub init {
         loop_context_vars      => 0,
         case_sensitive         => $CASE_SENSITIVE_DEFAULT,
         debug                  => $DEBUG_DEFAULT,
+        debug_file             => 0,
+        objects                => 'strict',
         out_fh                 => 0,
         global_vars            => 0,
         default_escape         => $DEFAULT_ESCAPE,
         default_path           => PATH_DEREF,
         use_query              => $DEFAULT_QUERY,
+        #use_expressions        => 0,
         use_perl               => 0,
         open_mode              => '',
+        no_includes            => 0,
         %args,
     );
     $self->set_loop_context(1) if $args{loop_context_vars};
@@ -829,6 +858,13 @@ sub init {
     $self->set_default_escape( $defaults{default_escape} );
     $self->set_default_path( $defaults{default_path} );
     $self->set_use_query( $defaults{use_query} );
+    #$self->set_use_expressions( $defaults{use_expressions} );
+    if ($defaults{use_expressions}) {
+        require HTML::Template::Compiled::Expr;
+    }
+    if ($defaults{open_mode}) {
+        $defaults{open_mode} =~ s/^[<>]//; # <:utf8
+    }
     $self->set_open_mode( $defaults{open_mode} );
     $self->set_search_path( $defaults{search_path_on_include} );
     $self->set_includes({});
@@ -838,6 +874,11 @@ sub init {
             HTML::Template::Compiled::Filter->new( $args{filter} ) );
     }
     $self->set_debug( $defaults{debug} );
+    $self->set_debug_file( $defaults{debug_file} );
+    $self->set_objects( $defaults{objects} );
+    if ($defaults{objects} and $defaults{objects} eq 'nostrict') {
+        require Scalar::Util;
+    }
     $self->set_out_fh( $defaults{out_fh} );
     $self->set_global_vars( $defaults{global_vars} );
     my $tagstyle = $args{tagstyle};
@@ -846,6 +887,7 @@ sub init {
         # user specified named styles or regexes
         $parser = $self->parser_class->new(
             tagstyle => $tagstyle,
+            use_expressions => $defaults{use_expressions},
         );
         $parser->set_perl($defaults{use_perl});
     }
@@ -856,6 +898,7 @@ sub init {
     unless ($parser) {
         $parser ||= $self->parser_class->default();
         $parser->set_perl($defaults{use_perl});
+        $parser->set_expressions($defaults{use_expressions});
     }
     if ($defaults{use_perl}) {
         $parser->add_tagnames({
@@ -864,21 +907,24 @@ sub init {
             }
         });
     }
+    if ($defaults{no_includes}) {
+        $parser->remove_tags(qw/ INCLUDE INCLUDE_VAR INCLUDE_STRING /);
+    }
     $self->set_parser($parser);
     my $compiler = $self->compiler_class->new;
     $self->set_compiler($compiler);
     if ($defaults{plugin}) {
-        my @plugins = ref $defaults{plugin} eq 'ARRAY' ? @{ $defaults{plugin} } : $defaults{plugin};;
-        $self->init_plugins(@plugins);
-        $self->set_plugins(\@plugins);
+        my $plugins = ref $defaults{plugin} eq 'ARRAY' ? $defaults{plugin} : [$defaults{plugin}];
+        $self->init_plugins($plugins);
+        $self->set_plugins($plugins);
     }
 }
 
 sub init_plugins {
-    my ($self, @plugins) = @_;
+    my ($self, $plugins) = @_;
     my $parser = $self->get_parser;
     my $compiler = $self->get_compiler;
-    for my $plug (@plugins) {
+    for my $plug (@$plugins) {
         if (ref $plug) {
         }
         elsif ($plug =~ m/^::/) {
@@ -930,8 +976,8 @@ sub _readfile {
         open $fh, $file or die "Cannot open '$file': $!";
     }
     else {
-        $open_mode = '<' unless length $open_mode;
-        open $fh, $open_mode, $file or die "Cannot open '$file': $!";
+        $open_mode = '' unless length $open_mode;
+        open $fh, "<$open_mode", $file or die "Cannot open '$file': $!";
     }
     local $/;
     my $text = <$fh>;
@@ -1099,6 +1145,7 @@ sub new_from_object {
     $new->set_path($path);
     $new->set_perl(undef);
     if (my $cached = $new->from_cache) {
+        $cached->set_plugins($self->get_plugins);
         $cached->init_includes;
         return $cached
     }
@@ -1110,6 +1157,15 @@ sub new_from_object {
 sub prepare_for_cache {
     my ($self) = @_;
     $self->clear_params;
+    my @plugs = @{ $self->get_plugins || [] };
+    for my $i (0 .. $#plugs) {
+        if (ref $plugs[$i]) {
+            if ($plugs[$i]->can('serialize')) {
+                $plugs[$i] = $plugs[$i]->serialize();
+            }
+        }
+    }
+    $self->set_plugins(\@plugs);
 }
 
 sub preload {
@@ -1202,7 +1258,7 @@ sub query {
     # use query-like behaviour.
     return unless defined wantarray();
     #print STDERR "query(@_)\n";
-    my $info = $self->get_use_query
+    my $info = $self->get_parse_tree
         or do {
             $self->_error_no_query();
             return;
@@ -1217,7 +1273,7 @@ sub query {
     $tags = [$tags] unless ref $tags eq 'ARRAY';
     my $includes = $self->get_includes;
     my %include_info = map {
-        $includes->{$_}->[1] => $includes->{$_}->[2]->get_use_query;
+        $includes->{$_}->[1] => $includes->{$_}->[2]->get_parse_tree;
     } keys %{ $includes };
     for my $tag (@$tags) {
         my $value;
@@ -1308,7 +1364,21 @@ sub output {
         : $p;
     my $f = $self->get_file;
     $fh = \*STDOUT unless $fh;
-    $self->get_perl()->( $self, $p, \$p, $fh );
+    if ($DEBUG) {
+        my $output;
+        eval {
+            $output = $self->get_perl()->( $self, $p, \$p, $fh );
+        };
+        if ($@) {
+            $LAST_EXCEPTION = $@;
+            my $filename = $self->get_file;
+            die "Error while executing '$filename': $@";
+        }
+        return $output;
+    }
+    else {
+        $self->get_perl()->( $self, $p, \$p, $fh );
+    }
 }
 
 sub import {
@@ -1410,6 +1480,53 @@ sub popGlobalstack {
     }
 }
 
+sub debug_code {
+    my ($self, $html) = @_;
+    my $perl = $self->get_perl;
+    require B::Deparse;
+    my $deparse = B::Deparse->new("-p", "-sC");
+    my $body = $deparse->coderef2text($perl);
+    my $filename = $self->get_file;
+    #warn __PACKAGE__.':'.__LINE__.$".Data::Dumper->Dump([\$body], ['body']);
+    my $message = '';
+    if ($LAST_EXCEPTION and $LAST_EXCEPTION =~ m/at \(eval \d*\) line (\d+)\./) {
+        my $rline = $1;
+        my $line = $rline;
+        $line--;
+        my @lines = split m#$/#, $body;
+        if ($line > $#lines) {
+            $line = $#lines;
+        }
+        my $pre = $line > 0 ? join $/, @lines[0 .. $line - 1] : '';
+        my $post = $line < $#lines ? join $/, @lines[$line + 1 .. $#lines] : '';
+        my $error = "$/$/# ------------------- ERROR line $rline in template $filename -----------------$/";
+        my $last = $LAST_EXCEPTION;
+        $LAST_EXCEPTION =~ s#$/# #g;
+        $error .= "# $last$/$lines[$line]$/";
+        if ($html) {
+            for ($pre, $error, $post) {
+                s/</&lt;/g;
+                s/>/&gt;/g;
+            }
+            $message = <<"EOM";
+<table border="0" style="background-color: #eeeeee;"><tr><td><pre>$pre</pre></td></tr>
+<tr><td style="background-color: #ffffff; color: #ff0000"><pre>$error</pre></td></tr>
+<tr><td><pre>$post</pre></td></tr></table>
+EOM
+        }
+        else {
+            $message .= $pre;
+            $message .= $error;
+            $message .= $post;
+        }
+    }
+    else {
+        $message = $LAST_EXCEPTION;
+    }
+    return $message;
+
+}
+
 my $version_pod = <<'=cut';
 =pod
 
@@ -1419,7 +1536,7 @@ HTML::Template::Compiled - Template System Compiles HTML::Template files to Perl
 
 =head1 VERSION
 
-$VERSION = "0.90"
+$VERSION = "0.92"
 
 =cut
 
@@ -1589,6 +1706,19 @@ Additional loop variable (C<__counter__ -1>)
 
 Additional loop variable (see L<"TMPL_LOOP">)
 
+=item C<__filename__>, C<__filenameshort__> (since 0.91_001)
+
+Insert the template filename for debugging:
+
+    <%= __filename__ %>
+    <%= __filenameshort__ %>
+
+will turn out as:
+    templates/path/file.html
+    path/file.html
+
+See also option debug_file in L<"OPTIONS"> for adding the filename globally.
+
 =item TMPL_SWITCH, TMPL_CASE
 
 see L<"TMPL_SWITCH">
@@ -1742,6 +1872,8 @@ If you want to have your templates read in utf-8, use
 
 as an option.
 
+In the previous version, it was '<:utf8'. This is deprecated.
+
 =back
 
 To be compatible in all of the above options all use:
@@ -2263,6 +2395,42 @@ browse through the stack.
 Now everything will be escaped for HTML unless you explicitly specify C<ESCAPE=0> (no escaping)
 or C<ESCAPE=URL>.
 
+=item no_includes (since 0.92)
+
+Default is 0. If set to 1, the tags INCLUDE, INCLUDE_VAR and INCLUDE_STRING
+will cause a template syntax error when creating. This can be useful when opening
+untrusted templates, otherwise any file in the filesystem could be opened.
+
+=item debug_file (fixed) (since 0.91_001)
+
+Additionally to the context_vars __filename__ and __filenameshort__ you
+can enable filename debugging globally.
+
+If the option is set to 'start', at the start of every template will be added:
+    <!-- start templates/path/filename.html -->
+
+If set to 'end', at the end will be added:
+    <!-- end templates/path/filename.html -->
+
+If set to 'start,end', both coments will be added.
+
+If set to 'start,short', 'end,short' or 'start,end,short' the path
+to the templates will be stripped:
+    <!-- start path/filename.html -->
+    <!-- end path/filename.html -->
+
+=item objects (fixed) (since 0.91_001)
+
+if set to true, you can use method calls like
+    <%= object.method %>
+
+Default is 'strict' (true).
+If set to 'strict', the method will be called if we have an object, otherwise
+it's treated as a hash lookup. If the method doesn't exist, it dies.
+If set to 'nostrict', the method will be called only if the object 'can' do the
+method, otherwise it will return undef (this will need Scalar::Util).
+If set to 0, no method calls are allowed.
+
 =item deref (fixed)
 
 Deprecated. Please inherit and overwrite method 'deref'. See L<"INHERITANCE">
@@ -2423,6 +2591,28 @@ tt-style and your own C<{{}} > style, then say:
         ],
     );
 
+=item use_expressions (since 0.91_003)
+
+Set to 1 if you want to use expressions. They work more or less like
+in L<HTML::Template::Expr> - I took the parsing code from it and
+used it with some minor changes - thanks to Sam Tregar.
+
+    <%if expr="some.var > 3" %>It's grater than 3<%/if %>
+
+Additionally you can use object methods with parameters. While a
+normal method call can only be called without parameters, like
+
+    <%= object.name %>
+
+with expressions you can give it parameters:
+
+    <%= expr="object.create_link('navi')" %>
+
+Inside function and method calls you also can use template
+vars.
+
+It is only minimally tested yet, so use with care and please report any
+bugs you find.
 
 =item formatter
 
@@ -2529,6 +2719,29 @@ cache directory. See L<"PRECOMPILE">.
 
 Empty all parameters.
 
+=item debug_code (since 0.91_003)
+
+If you get an error from the generated template, you might want to debug
+the executed code. You can now call C<debug_code> to get the compiled code
+and the line the error occurred. Note that the reported line might not be
+the exact line where the error occurred, also look around the line.
+The template filename reported does currently only report the main template,
+not the name of an included template. I'll try to fix that.
+
+    local $HTML::Template::Compiled::DEBUG = 1;
+    my $htc = HTML::Template::Compiled->new(
+        filename => 'some_file_with_runtime_error.html',
+    );
+    eval {
+        print $htc->output;
+    };
+    if ($@) {
+        # reports as text
+        my $msg = $htc->debug_code;
+        # reports as a html table
+        my $msg_html = $htc->debug_code('html');
+    }
+
 =item get_plugin
 
     my $plugin = $htc->get_plugin('Name::of::plugin');
@@ -2717,7 +2930,7 @@ Bjoern Kriews for original idea and contributions
 Special Thanks to Sascha Kiefer - he finds all the bugs!
 
 Ronnie Neumann, Martin Fabiani, Kai Sengpiel, Jan Willamowius, Justin Day,
-Steffen Winkler for ideas, beta-testing and patches
+Steffen Winkler, Henrik Tougaard for ideas, beta-testing and patches
 
 perlmonks.org and perl-community.de for everyday learning
 
